from django.test import TestCase

from mock import patch
from money.Money import Money

from .. import settings
from ..models import Payment
from ..api.datatypes import Receiver, ReceiverList


class MockPaymentRequest(object):
    _response = None
    _code = None

    @property
    def response(self):
        return self._response

    @property
    def code(self):
        return self._code

    def call(self, *args, **kwargs):
        return self


class TestPaymentResponses(TestCase):
    def setUp(self):
        self.payment = Payment()
        self.payment.money = Money(100, "USD")
        self.payment.save()
        self.payment = Payment.objects.all()[0]

        george = Receiver(amount=10, email=settings.settings.PAYPAL_EMAIL,
                          primary=False)
        allen = Receiver(amount=90, email="mrbuyer@antonagestam.se",
                         primary=True)
        self.receivers = ReceiverList([george, allen])

    @patch("paypaladaptive.api.endpoints.UrlRequest", MockPaymentRequest)
    def testPaymentExecStatusError(self):
        MockPaymentRequest._response = (
            u'{"responseEnvelope":'
            u'{"timestamp":"2013-06-14T07:29:00.573-07:00",'
            u'"ack":"Success",'
            u'"correlationId":"0a2a453266f0c",'
            u'"build":"6133763"},'
            u'"payKey":"AP-7YC11928U7484393G",'
            u'"paymentExecStatus":"ERROR",'
            u'"payErrorList":'
            u'{"payError":[{'
            u'"receiver":{'
            u'"amount":"2000",'
            u'"email":"hi@example.com",'
            u'"primary":"true"},'
            u'"error":{'
            u'"errorId":"569059",'
            u'"domain":"PLATFORM",'
            u'"severity":"Error",'
            u'"category":"Application",'
            u'"message":"Instant payments can\'t be pending"}}]}}')

        res = self.payment.process(self.receivers)
        self.assertFalse(res)
        self.assertEqual(self.payment.status, 'error')
        self.assertEqual(
            self.payment.status_detail,
            u"Error 569059: Instant payments can\'t be pending")



import logging
import  urllib

from django.utils import simplejson

from dateutil.parser import parse
from money.Money import Money, Currency
from pytz import utc

from constants import *
from paypaladaptive import settings
from paypaladaptive.api.errors import IpnError
from paypaladaptive.api.httpwrapper import UrlRequest


logger = logging.getLogger(__name__)


class IPN(object):
    """
    Models the IPN API response
    
    Note that this model is specific to the Paypal Adaptive API; it cannot
    handle IPNs from the standard Paypal checkout.

    """

    class Transaction(object):
        def __init__(self, **kwargs):
            self.id = kwargs.get('id', None)
            self.status = kwargs.get('status', None)
            self.id_for_sender = kwargs.get('id_for_sender', None)
            self.status_for_sender_txn = kwargs.get('status_for_sender_txn',
                                                    None)
            self.refund_id = kwargs.get('refund_id', None)
            self.refund_amount = IPN.process_money(kwargs.get('refund_amount',
                                                              None))
            self.refund_account_charged = kwargs.get('refund_account_charged',
                                                     None)
            self.receiver = kwargs.get('receiver', None)
            self.invoiceId = kwargs.get('invoiceId', None)            
            self.amount = IPN.process_money(kwargs.get('amount', None))
            self.is_primary_receiver = (kwargs.get('is_primary_receiver', '')
                                        == 'true')
            
        @classmethod
        def slicedict(cls, d, s):
            """
            Iterates over a dict d and filters out all values whose key starts
            with a string s and then removes that string s from the key and
            returns a new dict.

            """

            d = dict((str(k.replace(s, '', 1)), v) for k,v in d.iteritems() if k.startswith(s))
            return d

    def __init__(self, request):
        # verify that the request is paypal's
        url = '%s?cmd=_notify-validate' % settings.PAYPAL_PAYMENT_HOST
        post_data = {}
        for k, v in request.POST.copy().iteritems():
            post_data[k] = unicode(v).encode('utf-8')
        data = urllib.urlencode(post_data)
        verify_request = UrlRequest().call(url, data=data)

        # check code
        if verify_request.code != 200:
            raise IpnError('PayPal response code was %s' % verify_request.code)

        # check response
        raw_response = verify_request.response
        if raw_response != 'VERIFIED':
            raise IpnError('PayPal response was "%s"' % raw_response)

        # check transaction type
        raw_type = request.POST.get('transaction_type', '')
        allowed_types = [IPN_TYPE_PAYMENT, IPN_TYPE_ADJUSTMENT,
                         IPN_TYPE_PREAPPROVAL]

        if raw_type in allowed_types:
            self.type = raw_type
        else:
            raise IpnError('Unknown transaction_type received: %s' % raw_type)
        
        self.process_transactions(request)

        try:
            # payments and adjustments define these
            self.status = request.POST.get('status', None)
            self.sender_email = request.POST.get('sender_email', None)
            self.action_type = request.POST.get('action_type', None)
            self.payment_request_date = IPN.process_date(request.POST.get('payment_request_date', None))
            self.reverse_all_parallel_payments_on_error = request.POST.get('reverse_all_parallel_payments_on_error', 'false') == 'true'
            self.return_url = request.POST.get('return_url', None)
            self.cancel_url = request.POST.get('cancel_url', None)
            self.ipn_notification_url = request.POST.get('ipn_notification_url', None)
            self.pay_key = request.POST.get('pay_key', None)
            self.memo = request.POST.get('memo', None)
            self.fees_payer = request.POST.get('fees_payer', None) 
            self.trackingId = request.POST.get('trackingId', None)
            self.preapproval_key = request.POST.get('preapproval_key', None)
            self.reason_code = request.POST.get('reason_code', None)

            # preapprovals define these
            self.approved = request.POST.get('approved', 'false') == 'true'            
            self.current_number_of_payments = IPN.process_int(request.POST.get('current_number_of_payments', None))
            self.current_total_amount_of_all_payments = IPN.process_money(request.POST.get('current_total_amount_of_all_payments', None))
            self.current_period_attempts = IPN.process_int(request.POST.get('current_period_attempts', None))
            self.currency_code = Currency(request.POST.get('currency_code', None))
            self.date_of_month = IPN.process_int(request.POST.get('date_of_month', None))
            self.day_of_week = IPN.process_int(request.POST.get('day_of_week', None), None)
            self.starting_date = IPN.process_date(request.POST.get('starting_date', None))
            self.ending_date = IPN.process_date(request.POST.get('ending_date', None))
            self.max_total_amount_of_all_payments = Money(request.POST.get('max_total_amount_of_all_payments', None),
                                                          request.POST.get('currency_code', None))
            self.max_amount_per_payment = IPN.process_money(request.POST.get('max_amount_per_payment', None))
            self.max_number_of_payments = IPN.process_int(request.POST.get('max_number_of_payments', None))
            self.payment_period = request.POST.get('payment_period', None)
            self.pin_type = request.POST.get('pin_type', None)
        except Exception, e:
            logger.error('Could not parse request')
            raise e
        
        # Verify enumerations
        allowed_statuses = [IPN_STATUS_CREATED,
                            IPN_STATUS_COMPLETED,
                            IPN_STATUS_INCOMPLETE,
                            IPN_STATUS_ERROR,
                            IPN_STATUS_REVERSALERROR,
                            IPN_STATUS_PROCESSING,
                            IPN_STATUS_PENDING,
                            IPN_STATUS_ACTIVE,
                            IPN_STATUS_CANCELED]

        if self.status and self.status not in allowed_statuses:
            raise IpnError("unknown status: %s" % self.status)
        
        if (self.action_type
            and self.action_type not in [IPN_ACTION_TYPE_PAY,
                                         IPN_ACTION_TYPE_CREATE]):
            raise IpnError("unknown action type: %s" % self.action_type)

    @classmethod
    def process_int(cls, int_str, default='null'):
        """
        Attempt to turn strings into integers (or longs if long enough.)
        Doesn't trap the ValueError so bad values raise the exception.

        """

        val = default

        if int_str:
            try:
                val = int(int_str)
            except ValueError, e:
                if default == 'null':
                    raise e

        
        return val
    
    @classmethod
    def process_money(cls, money_str):
        """
        Paypal sends money in the form "XXX 0.00" where XXX is the currency
        code and 0.00 is the amount

        """

        if money_str:
            money_args = str(money_str).split(' ', 1)
            money_args.reverse()
            if money_args and len(money_args) == 2:
                return Money(*money_args)
        
        return None

    @classmethod
    def process_date(cls, date_str):
        """
        Paypal sends dates in the form "Thu Jun 09 07:23:38 PDT 2011", where
        the timezone appears to always be (US) Pacific.

        """

        if not date_str:
            return None
        
        return parse(date_str, tzinfos=IPN_TIMEZONES).astimezone(utc)
    
    def process_transactions(self, request):
        """
        Paypal sends transactions in the form transaction[n].[attribute], where
        n is 0 - 5 inclusive. We'll attempt to pull POST dictionary keys
        matching each transaction and build an IPN. Transaction object from
        them.

        """

        self.transactions = []

        transaction_nums = range(6)
        for transaction_num in transaction_nums:
            transdict = IPN.Transaction.slicedict(request.POST, 'transaction[%s].' % transaction_num)
            if len(transdict) > 0:
                self.transactions.append(IPN.Transaction(**transdict))


"""Endpoints for (parts of) Paypal Adaptive API."""

from datetime import datetime, timedelta
import logging

from django.utils import simplejson as json

from money.Money import Money

from paypaladaptive import settings

from errors import *
from datatypes import ReceiverList
from httpwrapper import UrlRequest

logger = logging.getLogger(__name__)


class PaypalAdaptiveEndpoint(object):
    """Base class for all Paypal endpoints"""

    headers = {}
    data = {'requestEnvelope': {'errorLanguage': 'en_US'}, }
    raw_response = None
    response = None
    error_class = Exception
    url = None

    def __init__(self, *args, **kwargs):
        remote_address = kwargs.pop('remote_address', None)
        self._build_headers(remote_address=remote_address)
        self.data.update(self.prepare_data(*args, **kwargs))

    def _build_headers(self, remote_address=None):
        headers = {'X-PAYPAL-SECURITY-USERID': settings.PAYPAL_USERID,
                   'X-PAYPAL-SECURITY-PASSWORD': settings.PAYPAL_PASSWORD,
                   'X-PAYPAL-SECURITY-SIGNATURE': settings.PAYPAL_SIGNATURE,
                   'X-PAYPAL-REQUEST-DATA-FORMAT': 'JSON',
                   'X-PAYPAL-RESPONSE-DATA-FORMAT': 'JSON',
                   'X-PAYPAL-APPLICATION-ID': settings.PAYPAL_APPLICATION_ID}

        if remote_address:
            headers['X-PAYPAL-DEVICE-IPADDRESS'] = remote_address

        self.headers.update(headers)

    def call(self):
        request = UrlRequest().call(self.url, data=json.dumps(self.data),
                                    headers=self.headers)
        self.raw_response = request.response
        self.response = json.loads(request.response)

        logger.debug('headers are: %s' % str(self.headers))
        logger.debug('request is: %s' % str(self.data))
        logger.debug('response is: %s' % str(self.raw_response))

        if ('responseEnvelope' not in self.response
                or 'ack' not in self.response['responseEnvelope']
                or self.response['responseEnvelope']['ack']
                not in ['Success', 'SuccessWithWarning']):
            error_message = 'unknown'
            try:
                error_message = self.response['error'][0]['message']
            except KeyError:
                pass

            raise self.error_class(error_message)

    def prepare_data(self, *args, **kwargs):
        """
        Override this to set the correct data for the Endpoint. Has to return
        a dict.

        """

        raise NotImplementedError("Endpoint class needs to override the "
                                  "prepare_data method.")

    def pretty_response(self):
        print json.dumps(json.loads(self.raw_response), indent=4)

    def pretty_request(self):
        print json.dumps(self.data, indent=4)


class Pay(PaypalAdaptiveEndpoint):
    """Models the Pay API operation"""

    url = '%s%s' % (settings.PAYPAL_ENDPOINT, 'Pay')
    error_class = PayError

    def prepare_data(self, money, return_url, cancel_url, receivers,
                     ipn_url=None, **kwargs):
        """Prepare data for Pay API call"""

        if (not money or not isinstance(money, Money)
                or money <= Money('0.00', money.currency)):
            raise ValueError("amount must be a positive instance of Money")

        if (not isinstance(receivers, ReceiverList) or len(receivers) < 1):
            raise ValueError("receivers must be an instance of ReceiverList")
        
        data = {'actionType': 'PAY',
                'currencyCode': money.currency.code,
                'returnUrl': return_url,
                'cancelUrl': cancel_url}
        
        receiverList = {'receiver': receivers.to_dict()}
        data.update({'receiverList': receiverList})

        if ipn_url:
            data.update({'ipnNotificationUrl': ipn_url})

        if kwargs:
            data.update(**kwargs)

        return data

    @property
    def status(self):
        return self.response.get('paymentExecStatus', None)

    @property
    def paykey(self):
        return self.response.get('payKey', None)


class PaymentDetails(PaypalAdaptiveEndpoint):
    """
    Models the PaymentDetails API operation.
    Use this to retrieve data about a Payment from Paypal

    """

    url = '%s%s' % (settings.PAYPAL_ENDPOINT, 'PaymentDetails')
    error_class = PaypalAdaptiveApiError

    def prepare_data(self, payKey=None, transactionId=None, trackingId=None):
        """Prepare data for PaymentDetails API call"""

        if not any([payKey, transactionId, trackingId]):
            raise self.error_class("You need to supply one of payKey, "
                                   "transactionId or trackingId for Paypal to "
                                   "identify the payment")

        data = {}

        if payKey is not None:
            data.update({'payKey': payKey})

        if transactionId is not None:
            data.update({'transactionId': transactionId})

        if trackingId is not None:
            data.update({'trackingId': trackingId})

        return data


class Refund(PaypalAdaptiveEndpoint):
    """
    Models the Refund API operation
    
    Currently only a full refund is supported.

    """

    url = '%s%s' % (settings.PAYPAL_ENDPOINT, 'Refund')
    error_class = RefundError

    def prepare_data(self, pay_key):
        if not pay_key:
            raise ValueError("a payKey must be provided")
        
        return {'payKey': pay_key}


class CancelPreapproval(PaypalAdaptiveEndpoint):
    """
    Models the Cancel Preapproval API operation
    
    Currently only a full refund is supported.

    """

    url = '%s%s' % (settings.PAYPAL_ENDPOINT, 'CancelPreapproval')
    error_class = CancelPreapprovalError

    def prepare_data(self, preapproval_key):
        if not preapproval_key:
            raise ValueError("must provide a preapprovalKey")
        
        return {'preapprovalKey': preapproval_key}


class Preapprove(PaypalAdaptiveEndpoint):
    """
    Models the Preapproval API operation
    
    Currently only a single payment with a simple date range is supported

    """

    url = '%s%s' % (settings.PAYPAL_ENDPOINT, 'Preapproval')
    error_class = PreapproveError

    def prepare_data(self, money, return_url, cancel_url, ipn_url=None,
                     starting_date=datetime.utcnow(),
                     ending_date=(datetime.utcnow() + timedelta(days=90)),
                     pin_type='NOT_REQUIRED', max_payments=1,
                     max_payments_per_period=1, **kwargs):

        data = {'currencyCode': money.currency.code,
                'returnUrl': return_url,
                'cancelUrl': cancel_url,
                'startingDate': starting_date.isoformat(),
                'endingDate': ending_date.isoformat(),
                'maxNumberOfPayments': max_payments,
                'maxNumberOfPaymentsPerPeriod': max_payments_per_period,
                'maxTotalAmountOfAllPayments': float(money.amount),
                'pinType': pin_type}

        if kwargs:
            data.update(**kwargs)

        if ipn_url:
            data['ipnNotificationUrl'] = ipn_url
        
        return data

    @property
    def preapprovalkey(self):
        return self.response.get('preapprovalKey', None)


class PreapprovalDetails(PaypalAdaptiveEndpoint):
    """
    Models the PreapprovalDetails API operation.
    Use this to retrieve data about a Preapproval from Paypal

    """

    url = '%s%s' % (settings.PAYPAL_ENDPOINT, 'PreapprovalDetails')
    error_class = PaypalAdaptiveApiError

    def prepare_data(self, preapprovalKey):
        """Prepare data for PreapprovalDetails API call"""
        return {'preapprovalKey': preapprovalKey}


class ShippingAddress(PaypalAdaptiveEndpoint):
    url = '%s%s' % (settings.PAYPAL_ENDPOINT, 'GetShippingAddresses')

    def prepare_data(self, paykey):
        return {'key': paykey}



import django.test as test
from django.contrib.sites.models import Site
from django.core.urlresolvers import reverse
from django.http import HttpRequest

from money.Money import Money
import mock
import urlparse

from paypaladaptive.api.ipn import IPN
from paypaladaptive.models import Payment, Preapproval
from paypaladaptive.api.errors import IpnError

from factories import PreapprovalFactory, PaymentFactory
from helpers import (MockIPNVerifyRequest,
                     MockIPNVerifyRequestFail,
                     MockIPNVerifyRequestInvalid,
                     MockIPNVerifyRequestInvalidCode,
                     mock_ipn_call)




class TestPaymentIPN(test.TestCase):
    def setUp(self):
        self.payment = PaymentFactory.create(status='created')

    @mock.patch('paypaladaptive.api.ipn.endpoints.UrlRequest',
                MockIPNVerifyRequest)
    def mock_ipn_call(self, data, url=None):
        c = test.Client()
        url = url if url is not None else self.payment.ipn_url

        return c.post(url, data=data)

    def get_valid_IPN_data(self, money):
        money = str(money)
        return {
            'status': 'COMPLETED',
            'transaction_type': 'Adaptive Payment PAY',
            'transaction[0].id': '1',
            'transaction[0].amount': money,
            'transaction[0].status': 'COMPLETED',
        }

    def testVerificationCall(self):
        """Test that the verification call is made with correct params"""

        data = self.get_valid_IPN_data(self.payment.money)
        self.mock_ipn_call(data)

        verification_data = urlparse.parse_qs(MockIPNVerifyRequest.data)

        for k, v in data.iteritems():
            self.assertEqual([v], verification_data[k])

    def get_payment(self):
        return Payment.objects.get(pk=self.payment.pk)

    def testPasses(self):
        """Test valid IPN call"""

        data = self.get_valid_IPN_data(self.payment.money)
        response = self.mock_ipn_call(data)

        self.assertEqual(response.status_code, 204)
        self.assertEqual(response.content, '')

        payment = self.get_payment()

        self.assertEqual(payment.status, 'completed')

    def testSequenceCalls(self):
        """Test two valid IPN calls being received in sequence"""

        self.testPasses()
        self.testPasses()

    def testMismatchedAmounts(self):
        """Test mismatching amounts"""

        wrong_amount = str(Money("1337.23", 'SEK'))

        data = {
            'status': 'COMPLETED',
            'transaction_type': 'Adaptive Payment PAY',
            'transaction[0].id': 1,
            'transaction[0].amount': wrong_amount,
            'transaction[0].status': 'COMPLETED',
            'transaction[1].id': 2,
            'transaction[1].amount': wrong_amount,
            'transaction[1].status': 'COMPLETED',
        }

        self.mock_ipn_call(data)

        payment = self.get_payment()

        self.assertEqual(payment.status, 'error')
        self.assertEqual(payment.status_detail,
                         "IPN amounts didn't match. Payment requested %s. "
                         "Payment made %s"
                         % (payment.money, wrong_amount))

    def testMismatchedUUID(self):
        """Test error response if invalid secret UUID"""

        current_site = Site.objects.get_current()
        incorrect_UUID = 'thisisnotthecorrectuuid'
        kwargs = {'object_id': self.payment.id,
                  'object_secret_uuid': incorrect_UUID}
        internal_url = reverse('paypal-adaptive-ipn', kwargs=kwargs)
        ipn_url = "http://%s%s" % (current_site, internal_url)

        money = str(self.payment.money)
        data = {
            'status': 'COMPLETED',
            'transaction_type': 'Adaptive Payment PAY',
            'transaction[0].id': '1',
            'transaction[0].amount': money,
            'transaction[0].status': 'COMPLETED',
        }

        response = self.mock_ipn_call(data, ipn_url)

        payment = self.get_payment()

        self.assertEqual(payment.status, 'error')
        self.assertEqual(payment.status_detail,
                         ('IPN secret "%s" did not match db'
                          % incorrect_UUID))

        self.assertEqual(response.status_code, 400)

    def test404MissingObject(self):
        """Test 404 if missing object"""

        current_site = Site.objects.get_current()
        incorrect_UUID = 'thisisnotthecorrectuuid'
        incorrect_object_id = 9000
        kwargs = {'object_id': incorrect_object_id,
                  'object_secret_uuid': incorrect_UUID}
        internal_url = reverse('paypal-adaptive-ipn', kwargs=kwargs)
        ipn_url = "http://%s%s" % (current_site, internal_url)

        money = str(self.payment.money)
        data = {
            'status': 'COMPLETED',
            'transaction_type': 'Adaptive Payment PAY',
            'transaction[0].id': '1',
            'transaction[0].amount': money,
            'transaction[0].status': 'COMPLETED',
        }

        response = self.mock_ipn_call(data, ipn_url)
        self.assertEqual(response.status_code, 404)


class TestPreapprovalIPN(test.TestCase):
    def setUp(self):
        self.preapproval = PreapprovalFactory.create(status='created')

    def mock_ipn_call(self, data, url=None):
        url = url if url is not None else self.preapproval.ipn_url
        return mock_ipn_call(data, url)

    def get_preapproval(self):
        return Preapproval.objects.get(pk=self.preapproval.pk)

    def get_valid_IPN_call(self, money):
        return {
            u'verify_sign': u'AFcWxV21C7fd0v3bYYYRCpSSRl31AzHvdpOG25tWUY92LjYc'
                            u'NpDPvi76',
            u'ending_date': u'2013-07-06T11:34:05.000-07:00',
            u'max_number_of_payments': u'1',
            u'payment_period': u'0',
            u'current_number_of_payments': u'0',
            u'sender_email': u'mrbuyer@antonagestam.se',
            u'current_total_amount_of_all_payments': u'0.00',
            u'starting_date': u'2013-05-08T16:04:24.000-07:00',
            u'notify_version': u'UNVERSIONED',
            u'preapproval_key': u'PA-94D908252K3063112',
            u'status': u'ACTIVE',
            u'date_of_month': u'0',
            u'day_of_week': u'NO_DAY_SPECIFIED',
            u'transaction_type': u'Adaptive Payment PREAPPROVAL',
            u'pin_type': u'NOT_REQUIRED',
            u'charset': u'windows-1252',
            u'test_ipn': u'1',
            u'currency_code': str(money.currency),
            u'current_period_attempts': u'0',
            u'max_total_amount_of_all_payments': str(money.amount),
            u'approved': u'true'
        }

    def testPasses(self):
        """Test valid IPN call"""

        money = self.preapproval.money
        data = self.get_valid_IPN_call(money)

        response = self.mock_ipn_call(data)
        self.assertEqual(response.status_code, 204)
        self.assertEqual(response.content, '')

        preapproval = self.get_preapproval()

        self.assertEqual(preapproval.status, 'approved')

    def testSequenceCalls(self):
        """Test two valid IPN calls being received in sequence"""

        self.testPasses()
        self.testPasses()

    @mock.patch('paypaladaptive.api.ipn.endpoints.UrlRequest',
                MockIPNVerifyRequest)
    def testVerificationCall(self):
        """Test that the verification call is made with correct params"""

        data = self.get_valid_IPN_call(self.preapproval.money)

        self.mock_ipn_call(data)

        verification_data = urlparse.parse_qs(MockIPNVerifyRequest.data)

        for k, v in data.iteritems():
            self.assertEqual([v], verification_data[k])

    def testMismatchedAmounts(self):
        """Test mismatching amounts"""

        wrong_amount = Money("1337.23", 'SEK')
        data = self.get_valid_IPN_call(wrong_amount)

        self.mock_ipn_call(data)

        preapproval = self.get_preapproval()

        self.assertEqual(preapproval.status, 'error')
        self.assertEqual(preapproval.status_detail,
                         ("IPN amounts didn't match. Preapproval "
                          "requested %s. Preapproval made %s"
                          % (preapproval.money, wrong_amount)))

    def testNotApproved(self):
        """Test unapproved"""

        data = self.get_valid_IPN_call(self.preapproval.money)
        data.update({u'approved': u'false'})

        self.mock_ipn_call(data)

        preapproval = self.get_preapproval()

        self.assertEqual(preapproval.status, 'error')
        self.assertEqual(preapproval.status_detail,
                         "The preapproval is not approved")

    def testUnicodeInMemo(self):
        """Test with Unicode characters in the memo field."""

        data = self.get_valid_IPN_call(self.preapproval.money)
        data.update({u'memo': u"v\ufffdrldens b\ufffdsta app"})

        self.mock_ipn_call(data)


class TestIPNVerification(test.TestCase):
    def setUp(self):
        self.request = HttpRequest()

    @mock.patch('paypaladaptive.api.ipn.endpoints.UrlRequest',
                MockIPNVerifyRequestFail)
    def testVerificationNoneCode(self):
        with self.assertRaises(IpnError) as context:
            IPN(self.request)

        self.assertEqual(context.exception.message,
                         'PayPal response code was None')

    @mock.patch('paypaladaptive.api.ipn.endpoints.UrlRequest',
                MockIPNVerifyRequestInvalidCode)
    def testVerificationCodeInvalid(self):
        with self.assertRaises(IpnError) as context:
            IPN(self.request)

        self.assertEqual(context.exception.message,
                         'PayPal response code was 500')

    @mock.patch('paypaladaptive.api.ipn.endpoints.UrlRequest',
                MockIPNVerifyRequestInvalid)
    def testVerificationMessageInvalid(self):
        with self.assertRaises(IpnError) as context:
            IPN(self.request)

        self.assertEqual(context.exception.message,
                         'PayPal response was "invalid"')



from django.test import TestCase
from django.test import Client

from ..models import Preapproval

from .factories import PreapprovalFactory
from .helpers import mock_ipn_call


class TestPreapprovalCancel(TestCase):
    def setUp(self):
        self.preapproval = PreapprovalFactory.create(status='created')

    def get_preapproval(self):
        return Preapproval.objects.get(pk=self.preapproval.pk)

    def hitCancelUrl(self):
        data = {'next': '/'}
        client = Client()
        return client.get(self.preapproval.cancel_url, data=data, follow=False)

    def get_valid_IPN_call(self, money):
        return {
            u'verify_sign': u'AFcWxV21C7fd0v3bYYYRCpSSRl31AzHvdpOG25tWUY92LjYc'
                            u'NpDPvi76',
            u'ending_date': u'2013-07-06T11:34:05.000-07:00',
            u'max_number_of_payments': u'1',
            u'payment_period': u'0',
            u'current_number_of_payments': u'0',
            u'sender_email': u'mrbuyer@antonagestam.se',
            u'current_total_amount_of_all_payments': u'0.00',
            u'starting_date': u'2013-05-08T16:04:24.000-07:00',
            u'notify_version': u'UNVERSIONED',
            u'preapproval_key': u'PA-94D908252K3063112',
            u'status': u'CANCELED',
            u'date_of_month': u'0',
            u'day_of_week': u'NO_DAY_SPECIFIED',
            u'transaction_type': u'Adaptive Payment PREAPPROVAL',
            u'pin_type': u'NOT_REQUIRED',
            u'charset': u'windows-1252',
            u'test_ipn': u'1',
            u'currency_code': str(money.currency),
            u'current_period_attempts': u'0',
            u'max_total_amount_of_all_payments': str(money.amount),
            u'approved': u'false'
        }

    def testPassing(self):
        """Test cancel URL does what it should"""

        self.preapproval.status = 'created'
        self.preapproval.preapproval_key = 'PA-9HW83863H61516232'
        self.preapproval.save()

        response = self.hitCancelUrl()
        # Accept normal response or redirect
        self.assertIn(response.status_code, [200, 302])

        preapproval = self.get_preapproval()

        self.assertEqual(preapproval.status_detail, '')
        self.assertEqual(preapproval.status, 'created')

    def testIPN(self):
        """Test IPN call does what it should"""
        response = mock_ipn_call(
            data = self.get_valid_IPN_call(self.preapproval.money),
            url = self.preapproval.ipn_url)

        self.assertEqual(response.status_code, 204)
        self.assertEqual(response.content, '')

        preapproval = self.get_preapproval()
        self.assertEqual(preapproval.status_detail,
                         'Cancellation received via IPN')
        self.assertEqual(preapproval.status, 'canceled')



from django.test import TestCase

from money.Money import Money
from mock import patch

from paypaladaptive import settings
from paypaladaptive.models import Preapproval, Payment
from paypaladaptive.api import Receiver, ReceiverList
from paypaladaptive.api.httpwrapper import UrlResponse

from mock_url_request import MockUrlRequest
from factories import PreapprovalFactory


if not settings.TEST_WITH_MOCK:
    class FakePatch:
        def __init__(self, *args, **kwargs):
            pass

        def __call__(self, f):
            return f

    patch = FakePatch


class MockUrlRequestPreapproval(MockUrlRequest):

    def _set_response(self):
        data = """{\"responseEnvelope\":
                   {\"timestamp\": \"2013-04-23T02:03:21.737-07:00\",
                    \"ack\": \"Success\",
                    \"correlationId\": \"877ca0ec3b8db\",
                    \"build\": \"5710487\"},
                   \"preapprovalKey\": \"PA-8X048594NU8990615\"}"""
        self._response = UrlResponse(data, {}, 200)
        return self

    def _endpoint_name(self):
        return "Preapproval"

    def _assert_valid_data(self, data):
        super(MockUrlRequestPreapproval, self)._assert_valid_data(data)
        assert data["currencyCode"] == str(self.preapproval().currency)
        assert data["maxNumberOfPayments"] == 1
        assert data["maxNumberOfPaymentsPerPeriod"] == 1
        assert data["endingDate"] is not None
        assert data["startingDate"] is not None
        assert data["pinType"] == "NOT_REQUIRED"
        assert data["maxTotalAmountOfAllPayments"] == self.preapproval().amount

    def _assert_valid_return_url(self, url):
        kwargs = {'preapproval_id': self.preapproval().id,
                  'secret_uuid': self.preapproval().secret_uuid}
        self._assert_valid_reversed_url(url, kwargs,
                                        'paypal-adaptive-preapproval-return')

    def _assert_valid_cancel_url(self, url):
        kwargs = {'preapproval_id': self.preapproval().id}
        self._assert_valid_reversed_url(url, kwargs,
                                        'paypal-adaptive-preapproval-cancel')


class MockUrlRequestPayment(MockUrlRequest):
    def _set_response(self):
        data = """{\"responseEnvelope\":
                   {\"timestamp\":\"2013-04-23T07:45:36.456-07:00\",
                    \"ack\":\"Success\",
                    \"correlationId\":\"0d9ee9c8aab09\",
                    \"build\":\"5710487\"},
                   \"payKey\":\"AP-7KL74713BP0955948\",
                   \"paymentExecStatus\":\"CREATED\"}"""
        self._response = UrlResponse(data, {}, 200)

    def _endpoint_name(self):
        return "Pay"

    def _assert_valid_return_url(self, url):
        kwargs = {'payment_id': self.payment().id,
                  'secret_uuid': self.payment().secret_uuid}
        self._assert_valid_reversed_url(url, kwargs,
                                        'paypal-adaptive-payment-return')

    def _assert_valid_cancel_url(self, url):
        self._assert_valid_reversed_url(url, {'payment_id': self.payment().id,
                                        'secret_uuid': self.payment().secret_uuid},
                                        'paypal-adaptive-payment-cancel')

    def _assert_valid_data(self, data):
        super(MockUrlRequestPayment, self)._assert_valid_data(data)
        assert data["currencyCode"] == str(self.payment().currency)
        assert data["actionType"] == "PAY"
        self._assert_valid_recievers_list(data["receiverList"]["receiver"])

    def _assert_valid_recievers_list(self, reciever_list):
        assert reciever_list == self.recievers().to_dict()


class FakePay:
    status = 'COMPLETED'
    paykey = 'PA-123'


class AdaptiveTests(TestCase):

    def setUp(self):

        self.preapproval = Preapproval()
        self.preapproval.money = Money(100, "SEK")
        self.preapproval.save()
        self.preapproval = Preapproval.objects.all()[0]

        self.payment = Payment()
        self.payment.money = Money(100, "USD")
        self.payment.save()
        self.payment = Payment.objects.all()[0]

        george = Receiver(amount=10, email=settings.settings.PAYPAL_EMAIL,
                          primary=False)
        allen = Receiver(amount=90, email="mrbuyer@antonagestam.se",
                         primary=True)
        self.recievers = ReceiverList([george, allen])

    @patch("paypaladaptive.api.endpoints.UrlRequest",
           MockUrlRequestPreapproval)
    def test_preapproval(self):
        MockUrlRequestPreapproval.preapproval = (
            classmethod(lambda cls: self.preapproval))
        self.assertTrue(self.preapproval.process())
        self.preapproval = Preapproval.objects.get(pk=self.preapproval.pk)
        self.assertNotEqual(self.preapproval.preapproval_key, "")
        self.assertEqual(self.preapproval.status, "created")

    @patch("paypaladaptive.api.endpoints.UrlRequest",
           MockUrlRequestPayment)
    def test_payment(self):
        MockUrlRequestPayment.payment = classmethod(lambda cls: self.payment)
        MockUrlRequestPayment.recievers = classmethod(
            lambda cls: self.recievers)
        self.assertTrue(self.payment.process(self.recievers))
        self.payment = Payment.objects.get(pk=self.payment.id)
        self.assertTrue(self.payment.status, "created")
        self.assertNotEqual(self.payment.pay_key, "")

    @patch("paypaladaptive.models.Payment.call")
    def test_payment_from_preapproval(self, mock_api_caller):
        """Make sure we can create payments from Preapprovals"""

        preapproval = PreapprovalFactory.create(status='approved')
        mock_api_caller.return_value = (True, FakePay())

        self.assertTrue(self.payment.process(self.recievers,
                                             preapproval=self.preapproval))
        self.assertEqual(self.payment.status, 'completed')



'''
Base forms for user-interaction and payments

Created on Jun 14, 2011

@author: greg
'''
from django import forms
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _, to_locale, get_language
import settings

def paypal_image_url(type='pay'):
    '''
    Due to the way Paypal supports localization, we can't ask for a valid two character locale.
    So if our locale is such a one we'll have to modify it to fit Paypal's expectations.
    
    eg: 'en' -> 'en_US'
    '''
    current_locale = to_locale(get_language())
    if current_locale.find('_') == -1:
        current_locale = "%s_%s" % (current_locale, 
                                    current_locale.upper() if current_locale != 'en' else 'US')

    if type == 'pay':
        return 'https://www.paypal.com/%s/i/btn/btn_dg_pay_w_paypal.gif' % current_locale
    
    raise ValueError('Unknown image type')
 
    
class PayPalAdaptiveEmbeddedForm(forms.Form):
    """
    Form used to provide access to an embedded checkout form from Paypal.
    
    You must provide the form with a valid paykey from the Pay API operation.
    """
    expType = forms.CharField(initial='light', widget=forms.HiddenInput)
    payKey = forms.CharField(widget=forms.HiddenInput)

    def __init__(self, paykey, submit_title=_(u'Pay with Paypal'), *args, **kwargs):
        super(PayPalAdaptiveEmbeddedForm, self).__init__(*args, **kwargs)
        self.auto_id = '%s'
        self.initial['payKey'] = paykey
        self.submit_title = submit_title

    def render(self):
        
        return mark_safe(u"""
            <form action="%(action)s" target="PPDGFrame">
                %(form)s
                <input type="image" id="submitBtn" src="%(image_url)s" alt="%(submit_title)s" />
            </form>
            <script type="text/javascript">
                var dgFlow;
                window.onload = function() {
                    dgFlow = new PAYPAL.apps.DGFlow({ trigger: 'submitBtn' });
                } 
            </script>
            """ % {'action': settings.EMBEDDED_ENDPOINT,
                   'form': self.as_p(),
                   'image_url': paypal_image_url(),
                   'submit_title': self.submit_title})

    class Media:
        js = ('http://www.paypalobjects.com/js/external/dg.js',)
        
class PayPalAdaptiveEmbeddedPreapprovalForm(forms.Form):
    """
    Form used to provide access to an embedded checkout form from Paypal.
    
    You must provide the form with a valid preapproval from the Pay API operation.
    """
    expType = forms.CharField(initial='light', widget=forms.HiddenInput)
    preapprovalKey = forms.CharField(widget=forms.HiddenInput)
    _cmd = forms.CharField(widget=forms.HiddenInput)

    def __init__(self, preapprovalkey, submit_title=_(u'Pay with Paypal'), *args, **kwargs):
        super(PayPalAdaptiveEmbeddedPreapprovalForm, self).__init__(*args, **kwargs)
        self.auto_id = '%s'
        self.initial['preapprovalKey'] = preapprovalkey
        self.submit_title = submit_title
        self.initial['_cmd'] = "_ap-preapproval"

    def render(self):
        
        return mark_safe(u"""
            <form action="%(action)s" target="PPDGFrame">
                %(form)s
                <input type="image" id="submitBtn" src="%(image_url)s" alt="%(submit_title)s" />
            </form>
            <script type="text/javascript">
                var dgFlow;
                window.onload = function() {
                    dgFlow = new PAYPAL.apps.DGFlow({ trigger: 'submitBtn' });
                } 
            </script>
            """ % {'action': settings.EMBEDDED_ENDPOINT,
                   'form': self.as_p(),
                   'image_url': paypal_image_url(),
                   'submit_title': self.submit_title})

    class Media:
        js = ('http://www.paypalobjects.com/js/external/dg.js',)


from django.test import TestCase
from django.test import Client

from paypaladaptive.models import Payment
from .factories import PaymentFactory


class TestPaymentReturnURL(TestCase):
    def setUp(self):
        self.payment = PaymentFactory.create(status='created')
        self.client = Client()

    def get_payment(self):
        return Payment.objects.get(pk=self.payment.pk)

    def hitReturnUrl(self, url=None):
        data = {'next': '/'}
        url = self.payment.return_url if url is None else url
        return self.client.get(url, data=data, follow=False)

    def testPassing(self):
        """Test that during normal circumstances, everything is OK"""

        response = self.hitReturnUrl()
        payment = self.get_payment()

        self.assertEqual(payment.status_detail, '')
        self.assertEqual(payment.status, 'returned')
        self.assertIn(response.status_code, [200, 302])

    def testAlreadyCompleted(self):
        """Test that 'completed' status is not overridden"""

        self.payment.status = 'completed'
        self.payment.save()

        response = self.hitReturnUrl()
        payment = self.get_payment()

        self.assertEqual(payment.status_detail, '')
        self.assertEqual(payment.status, 'completed')
        self.assertIn(response.status_code, [200, 302])

    def testUnexpectedStatus(self):
        """
        Test that receiving IPN for anything else than completed/created
        saves an error on the Payment.

        """

        self.payment.status = 'new'
        self.payment.save()

        response = self.hitReturnUrl()
        payment = self.get_payment()

        self.assertEqual(response.status_code, 500)
        self.assertEqual(payment.status, 'error')
        self.assertEqual(payment.status_detail,
                         u"Expected status to be created or completed, not "
                         u"new - duplicate transaction?")



"""Models to support Paypal Adaptive API"""
from datetime import datetime, timedelta

from django.core.urlresolvers import reverse
from django.db import models, transaction
from django.utils.translation import ugettext_lazy as _
from django.contrib.sites.models import Site
from django.utils import simplejson as json

from money.contrib.django.models.fields import MoneyField

import settings
import api


try:
    import uuid
except ImportError:
    from django.utils import uuid


class UUIDField(models.CharField) :
    """Django db field using python's uuid4 library"""

    def __init__(self, *args, **kwargs):
        kwargs['max_length'] = kwargs.get('max_length', 32)
        models.CharField.__init__(self, *args, **kwargs)
    
    def pre_save(self, model_instance, add):
        if add:
            value = getattr(model_instance, self.attname)
            if not value:
                value = unicode(uuid.uuid4().hex)
            setattr(model_instance, self.attname, value)
        else:
            value = super(models.CharField, self).pre_save(model_instance, add)

        return value


class PaypalAdaptive(models.Model):
    """Base fields used by all PaypalAdaptive models"""
    money = MoneyField(_(u'money'), max_digits=settings.MAX_DIGITS,
                       decimal_places=settings.DECIMAL_PLACES)
    created_date = models.DateTimeField(_(u'created on'), auto_now_add=True)
    secret_uuid = UUIDField(_(u'secret UUID'))  # to verify return_url
    debug_request = models.TextField(_(u'raw request'), blank=True, null=True)
    debug_response = models.TextField(_(u'raw response'), blank=True,
                                      null=True)

    def call(self, endpoint_class, *args, **kwargs):
        endpoint = endpoint_class(*args, **kwargs)

        try:
            res = endpoint.call()
        finally:
            self.debug_request = json.dumps(endpoint.data)
            self.debug_response = endpoint.raw_response
            self.save()

        return res, endpoint

    def get_amount(self):
        return self.money.amount

    def set_amount(self, value):
        self.money.amount = value

    amount = property(get_amount, set_amount)

    def get_currency(self):
        return self.money.currency

    def set_currency(self, value):
        self.money.currency = value

    currency = property(get_currency, set_currency)

    @property
    def ipn_url(self):
        current_site = Site.objects.get_current()
        kwargs = {'object_id': self.id,
                  'object_secret_uuid': self.secret_uuid}
        ipn_url = reverse('paypal-adaptive-ipn', kwargs=kwargs)
        return "http://%s%s" % (current_site, ipn_url)
    
    class Meta:
        abstract = True

    def get_update_kwargs(self):
        return {}

    def _parse_update_status_detail(self, response):
        return ''

    def update(self, save=True, fields=None):
        if not hasattr(self, 'update_endpoint'):
            raise NotImplemented('Model need to specify an update endpoint')

        if fields is None:
            fields = ['status', 'status_detail']

        __, endpoint = self.call(self.update_endpoint,
                                 **self.get_update_kwargs())
        response = endpoint.response

        for field in fields:
            val = getattr(self, '_parse_update_%s' % field)(response)
            setattr(self, field, val)

        if save:
            self.save()


class Payment(PaypalAdaptive):
    """Models a payment made using Paypal"""

    update_endpoint = api.PaymentDetails

    STATUS_CHOICES = (
        ('new', _(u'New')),  # just saved locally
        ('created', _(u'Created')),  # payment created
        ('error', _(u'Error')),  # error occurred somewhere in the process
        ('canceled', _(u'Canceled')),  # the payment has been canceled
        ('returned', _(u'Returned')),  # user has returned via return_url
        ('completed', _(u'Completed')),  # the payment has been completed
        ('refunded', _(u'Refunded')),  # payment has been refunded
    )

    pay_key = models.CharField(_(u'paykey'), max_length=255)
    transaction_id = models.CharField(_(u'paypal transaction ID'),
                                      max_length=128, blank=True, null=True)
    status = models.CharField(_(u'status'), max_length=10,
                              choices=STATUS_CHOICES, default='new')
    status_detail = models.CharField(_(u'detailed status'), max_length=2048)

    def save(self, *args, **kwargs):
        is_new = self.id is None

        super(Payment, self).save(*args, **kwargs)

        if settings.USE_DELAYED_UPDATES and is_new:
            from .tasks import update_payment
            update_payment.apply_async(
                args=[self.id],
                eta=datetime.now() + settings.DELAYED_UPDATE_COUNTDOWN)

    @property
    def return_url(self):
        current_site = Site.objects.get_current()
        kwargs = {'payment_id': self.id, 'secret_uuid': self.secret_uuid}
        return_url = reverse('paypal-adaptive-payment-return', kwargs=kwargs)
        return "http://%s%s" % (current_site, return_url)

    @property
    def cancel_url(self):
        current_site = Site.objects.get_current()
        kwargs = {'payment_id': self.id, 'secret_uuid': self.secret_uuid}
        cancel_url = reverse('paypal-adaptive-payment-cancel', kwargs=kwargs)
        return "http://%s%s" % (current_site, cancel_url)

    @transaction.autocommit
    def process(self, receivers, preapproval=None, **kwargs):
        """Process the payment"""

        endpoint_kwargs = {'money': self.money,
                           'return_url': self.return_url,
                           'cancel_url': self.cancel_url}

        # Update return_url with ?next param
        if 'next' in kwargs:
            return_next = "%s?next=%s" % (self.return_url, kwargs.pop('next'))
            endpoint_kwargs.update({'return_url': return_next})

        # Update cancel_url
        if 'cancel' in kwargs:
            return_cancel = "%s?next=%s" % (self.cancel_url,
                                            kwargs.pop('cancel'))
            endpoint_kwargs.update({'cancel_url': return_cancel})

        # Set ipn_url
        if settings.USE_IPN:
            endpoint_kwargs.update({'ipn_url': self.ipn_url})

        # Append extra arguments
        endpoint_kwargs.update(**kwargs)

        # Validate type of receivers and check ReceiverList has primary,
        # otherwise assign first
        if not isinstance(receivers, api.ReceiverList):
            raise ValueError("receivers must be an instance of "
                             "ReceiverList")
        elif not receivers.has_primary() and settings.USE_CHAIN:
            receivers.receivers[0].primary = True

        endpoint_kwargs.update({'receivers': receivers})

        # Use preapproval
        if preapproval is not None:
            if not isinstance(preapproval, Preapproval):
                raise ValueError("preapproval must be an instance of "
                                 "Preapproval")

            key = preapproval.preapproval_key
            endpoint_kwargs.update({'preapprovalKey': key})

        # Append extra arguments
        endpoint_kwargs.update(kwargs)

        # Call endpoint
        res, endpoint = self.call(api.Pay, **endpoint_kwargs)

        self.pay_key = endpoint.paykey

        if endpoint.status == 'ERROR':
            self.status = 'error'
            if 'payErrorList' in endpoint.response:
                if 'payError' in endpoint.response['payErrorList']:
                    payError = endpoint.response[
                        'payErrorList']['payError'][0]['error']
                    self.status_detail = "%s %s: %s" % (
                        payError['severity'],
                        payError['errorId'],
                        payError['message'])
                else:
                    self.status_detail = json.dumps(
                        endpoint.response.payErrorList)

        elif endpoint.status == 'COMPLETED':
            self.status = 'completed'
        elif endpoint.paykey or endpoint.status == 'CREATED':
            self.status = 'created'
        else:
            self.status = 'error'
            
        self.save()
        
        return self.status in ['created', 'completed']

    @transaction.autocommit
    def refund(self):
        """Refund this payment"""

        # TODO: flow should create a Refund object and call Refund.process()

        self.save()
        
        if self.status != 'completed':
            raise ValueError('Cannot refund a Payment until it is completed.')

        res, refund_call = self.call(api.Refund, self.pay_key)

        self.status = 'refunded'
        self.save()
    
        refund = Refund(payment=self,
                        debug_request=json.dumps(refund_call.data),
                        debug_response=refund_call.raw_response)
        refund.save()

    def get_update_kwargs(self):
        if self.pay_key is None:
            raise ValueError("Can't update unprocessed payments")
        return {'payKey': self.pay_key}

    def _parse_update_status(self, response):
        status = response.get('status', None)

        if status == 'COMPLETED':
            return 'completed'
        elif status == 'CREATED':
            return 'created'
        elif status == 'ERROR':
            return 'error'
        else:
            return self.status

    def next_url(self):
        return ('%s?cmd=_ap-payment&paykey=%s'
                % (settings.PAYPAL_PAYMENT_HOST, self.pay_key))
            
    def __unicode__(self):
        return self.pay_key


class Refund(PaypalAdaptive):
    """Models a refund make using Paypal"""

    STATUS_CHOICES = (
        ('new', _(u'New')), 
        ('created', _(u'Created')), 
        ('error', _(u'Error')), 
        ('canceled', _(u'Canceled')), 
        ('returned', _(u'Returned')),
        ('completed', _(u'Completed')),
    )

    payment = models.OneToOneField(Payment)
    status = models.CharField(_(u'status'), max_length=10,
                              choices=STATUS_CHOICES, default='new')
    status_detail = models.CharField(_(u'detailed status'), max_length=2048)
    
    # TODO: finish model


class Preapproval(PaypalAdaptive):
    """Models a preapproval made using Paypal"""

    update_endpoint = api.PreapprovalDetails
    default_valid_range = timedelta(days=90)
    default_valid_date = lambda: (datetime.now() +
                                  Preapproval.default_valid_range)

    STATUS_CHOICES = (
        ('new', _(u'New')),
        ('created', _(u'Created')),
        ('error', _(u'Error')),
        ('canceled', _(u'Canceled')),
        ('approved', _(u'Approved')),
        ('used', _(u'Used')),
        ('returned', _(u'Returned')),
    )
    
    valid_until_date = models.DateTimeField(_(u'valid until'),
                                            default=default_valid_date)
    preapproval_key = models.CharField(_(u'preapprovalkey'), max_length=255)
    status = models.CharField(_(u'status'), max_length=10,
                              choices=STATUS_CHOICES, default='new')
    status_detail = models.CharField(_(u'detailed status'), max_length=2048)

    def save(self, *args, **kwargs):
        is_new = self.id is None

        super(Preapproval, self).save(*args, **kwargs)

        if settings.USE_DELAYED_UPDATES and is_new:
            from .tasks import update_preapproval
            update_preapproval.apply_async(
                args=[self.id],
                eta=datetime.now() + settings.DELAYED_UPDATE_COUNTDOWN)

    @property
    def return_url(self):
        current_site = Site.objects.get_current()
        kwargs = {'preapproval_id': self.id, 'secret_uuid': self.secret_uuid}
        return_url = reverse('paypal-adaptive-preapproval-return',
                             kwargs=kwargs)
        return "http://%s%s" % (current_site, return_url)

    @property
    def cancel_url(self):
        current_site = Site.objects.get_current()
        kwargs = {'preapproval_id': self.id}
        cancel_url = reverse('paypal-adaptive-preapproval-cancel',
                             kwargs=kwargs)
        return "http://%s%s" % (current_site, cancel_url)

    @transaction.autocommit
    def process(self, **kwargs):
        """Process the preapproval"""

        endpoint_kwargs = {'money': self.money,
                           'return_url': self.return_url,
                           'cancel_url': self.cancel_url,
                           'starting_date': self.created_date,
                           'ending_date': self.valid_until_date}

        if 'next' in kwargs:
            return_next = "%s?next=%s" % (self.return_url, kwargs.pop('next'))
            endpoint_kwargs.update({'return_url': return_next})

        if 'cancel' in kwargs:
            return_cancel = "%s?next=%s" % (self.cancel_url,
                                            kwargs.pop('cancel'))
            endpoint_kwargs.update({'cancel_url': return_cancel})

        if settings.USE_IPN:
            endpoint_kwargs.update({'ipn_url': self.ipn_url})

        # Append extra arguments
        endpoint_kwargs.update(**kwargs)

        res, preapprove = self.call(api.Preapprove, **endpoint_kwargs)
    
        if preapprove.preapprovalkey:
            self.preapproval_key = preapprove.preapprovalkey
            self.status = 'created'
        else:
            self.status = 'error'
            
        self.save()
        
        return self.status == 'created'
        
    @transaction.autocommit
    def cancel_preapproval(self):
        res, cancel = self.call(api.CancelPreapproval,
                                preapproval_key=self.preapproval_key)

        # TODO: validate response

        self.status = 'canceled'
        self.save()
        return self.status == 'canceled'
        
    @transaction.autocommit
    def mark_as_used(self):
        self.status = 'used'
        self.save()
        
        return self.status == 'used'

    def get_update_kwargs(self):
        if self.preapproval_key is None:
            raise ValueError("Can't update unprocessed preapprovals")
        return {'preapprovalKey': self.preapproval_key}

    def _parse_update_status(self, response):
        completed_payments = response.get('curPayments', None)
        max_payments = response.get('maxNumberOfPayments', None)
        status = response.get('status', None)
        approved = response.get('approved', 'false')

        if ('curPayments' in response and 'maxNumberOfPayments' in response
                and completed_payments == max_payments):
            return 'used'
        elif status == 'ACTIVE' and approved == 'true':
            return 'approved'
        elif status == 'ACTIVE':
            return 'created'
        elif status == 'CANCELED':
            return 'canceled'
        else:
            return self.status

    def next_url(self):
        """Custom next URL"""
        return ('%s?cmd=_ap-preapproval&preapprovalkey=%s'
                % (settings.PAYPAL_PAYMENT_HOST, self.preapproval_key))
            
    def __unicode__(self):
        return self.preapproval_key


try:
    from south.modelsinspector import add_introspection_rules
    # South support for the custom fields
    add_introspection_rules([], ["^paypaladaptive\.models\.UUIDField"])
    add_introspection_rules([], ["^paypaladaptive\.models\.MoneyField"])
except ImportError:
    # South is not installed
    pass



import urllib2


class UrlResponse:
    def __init__(self, data, meta, code):
        self.data = data
        self.meta = meta
        self.code = code


class UrlRequest:
    def call(self, url, data=None, headers=None):
        if headers is None:
            headers = {}

        request = urllib2.Request(url, data=data, headers=headers)

        try:
            response = urllib2.urlopen(request)

            self._response = UrlResponse(response.read(), response.info(),
                                         response.getcode())
        except urllib2.URLError, e:
            self._response = UrlResponse(e.reason, {}, None)

        return self

    @property
    def response(self):
        return self._response.data

    @property
    def code(self):
        return self._response.code



import json

from django.core.urlresolvers import reverse
from django.contrib.sites.models import Site

from paypaladaptive import settings


class MockUrlRequest(object):
    def call(self, url, data=None, headers=None):
        self._assert_valid_url(url)
        self._assert_valid_data(json.loads(data))
        self._assert_valid_headers(headers)
        self._set_response()
        return self

    def _set_response(self):
        """Define JSON string returned by paypal"""
        raise NotImplementedError

    def _assert_valid_url(self, url):
        valid_url = "%s%s" % (settings.PAYPAL_ENDPOINT, self._endpoint_name())
        assert url == valid_url

    def _endpoint_name(self):
        """"Name of API call to paypal"""
        raise NotImplementedError

    def _assert_valid_return_url(self, url):
        raise NotImplementedError

    def _assert_valid_cancel_url(self, url):
        raise NotImplementedError

    def _assert_valid_data(self, data):
        self._assert_valid_return_url(data["returnUrl"])
        self._assert_valid_cancel_url(data["cancelUrl"])
        assert data["requestEnvelope"] == {"errorLanguage": "en_US"}

    def _assert_valid_headers(self, headers):
        self._assert_valid_header(headers, 'X-PAYPAL-SECURITY-SIGNATURE',
                                  settings.PAYPAL_SIGNATURE)
        self._assert_valid_header(headers, 'X-PAYPAL-REQUEST-DATA-FORMAT',
                                  'JSON')
        self._assert_valid_header(headers, 'X-PAYPAL-APPLICATION-ID',
                                  settings.PAYPAL_APPLICATION_ID)
        self._assert_valid_header(headers, 'X-PAYPAL-SECURITY-USERID',
                                  settings.PAYPAL_USERID)
        self._assert_valid_header(headers, 'X-PAYPAL-SECURITY-PASSWORD',
                                  settings.PAYPAL_PASSWORD)
        self._assert_valid_header(headers, 'X-PAYPAL-RESPONSE-DATA-FORMAT',
                                  'JSON')

    def _assert_valid_header(self, headers, name, value):
        assert headers[name] == value

    def _assert_valid_reversed_url(self, url, kwargs, template_name):
        """"Helper function for return and cancel urls"""
        current_site = Site.objects.get_current()
        return_url = reverse(template_name, kwargs=kwargs)
        assert url == "http://%s%s" % (current_site, return_url)

    @property
    def response(self):
        return self._response.data

    @property
    def code(self):
        return self._response.code


from errors import ReceiverError


class Receiver():
    email = None
    amount = None
    primary = False

    def __init__(self, email=None, amount=None, primary=False):
        self.email = email
        self.amount = amount
        self.primary = primary

    def to_dict(self):
        return {'email': self.email,
                'amount': self.amount,
                'primary': self.primary}

    def __unicode__(self):
        return self.email


class ReceiverList():
    receivers = None

    def __init__(self, receivers=None):
        self.receivers = []
        if receivers is not None:
            for receiver in receivers:
                self.append(receiver)

    def append(self, receiver):
        if not isinstance(receiver, Receiver):
            raise ReceiverError("receiver needs to be instance of Receiver")
        self.receivers.append(receiver)

    def to_dict(self):
        self.has_primary()
        return [r.to_dict() for r in self.receivers]

    def __len__(self):
        return len(self.receivers)

    def has_primary(self):
        n_primary = len(filter(lambda r: r.primary is True, self.receivers))

        if n_primary > 1:
            raise ReceiverError("There can only be one primary Receiver")

        return n_primary == 1

    @property
    def total_amount(self):
        return sum([r.amount for r in self.receivers])


import django.test as test

from mock import patch

from ..api.httpwrapper import UrlRequest, UrlResponse

class MockIPNVerifyRequest(UrlRequest):
    def call(self, url, data=None, headers=None):
        MockIPNVerifyRequest.data = data
        self._response = UrlResponse(data='VERIFIED', meta={}, code=200)
        return self


class MockIPNVerifyRequestInvalid(UrlRequest):
    data = None
    def call(self, url, data=None, headers=None):
        self.data = data
        self._response = UrlResponse(data='invalid', meta={}, code=200)
        return self


class MockIPNVerifyRequestFail(UrlRequest):
    data = None
    def call(self, url, data=None, headers=None):
        self.data = data
        self._response = UrlResponse(data='invalid', meta={}, code=None)
        return self


class MockIPNVerifyRequestInvalidCode(UrlRequest):
    data = None
    def call(self, url, data=None, headers=None):
        self.data = data
        self._response = UrlResponse(data='VERIFIED', meta={}, code=500)
        return self

@patch('paypaladaptive.api.ipn.endpoints.UrlRequest', MockIPNVerifyRequest)
def mock_ipn_call(data, url):
    c = test.Client()
    return c.post(url, data=data)


"""
IPN constants

"""

from pytz import timezone

IPN_TYPE_PAYMENT = 'Adaptive Payment PAY'
IPN_TYPE_ADJUSTMENT = 'Adjustment'
IPN_TYPE_PREAPPROVAL = 'Adaptive Payment PREAPPROVAL'

IPN_STATUS_CREATED = 'CREATED'
IPN_STATUS_COMPLETED = 'COMPLETED'
IPN_STATUS_INCOMPLETE = 'INCOMPLETE'
IPN_STATUS_ERROR = 'ERROR' 
IPN_STATUS_REVERSALERROR = 'REVERSALERROR'
IPN_STATUS_PROCESSING = 'PROCESSING'
IPN_STATUS_PENDING = 'PENDING'
IPN_STATUS_ACTIVE = 'ACTIVE'
IPN_STATUS_CANCELED = 'CANCELED'

IPN_ACTION_TYPE_PAY = 'PAY'
IPN_ACTION_TYPE_CREATE = 'CREATE'

IPN_TXN_STATUS_COMPLETED = 'Completed'
IPN_TXN_STATUS_PENDING = 'Pending'
IPN_TXN_STATUS_REFUNDED = 'Refunded'

IPN_TXN_SENDER_STATUS_SUCCESS = 'SUCCESS' 
IPN_TXN_SENDER_STATUS_PENDING = 'PENDING'
IPN_TXN_SENDER_STATUS_CREATED = 'CREATED' 
IPN_TXN_SENDER_STATUS_PARTIALLY_REFUNDED = 'PARTIALLY_REFUNDED' 
IPN_TXN_SENDER_STATUS_DENIED = 'DENIED' 
IPN_TXN_SENDER_STATUS_PROCESSING = 'PROCESSING' 
IPN_TXN_SENDER_STATUS_REVERSED = 'REVERSED' 
IPN_TXN_SENDER_STATUS_REFUNDED = 'REFUNDED' 
IPN_TXN_SENDER_STATUS_FAILED = 'FAILED'

IPN_FEES_PAYER_SENDER = 'SENDER'
IPN_FEES_PAYER_PRIMARYRECEIVER = 'PRIMARYRECEIVER'
IPN_FEES_PAYER_EACHRECEIVER = 'EACHRECEIVER'
IPN_FEES_PAYER_SECONDARYONLY = 'SECONDARYONLY'

IPN_REASON_CODE_CHARGEBACK = 'Chargeback' 
IPN_REASON_CODE_SETTLEMENT = 'Settlement'
IPN_REASON_CODE_ADMIN_REVERSAL = 'Admin reversal'
IPN_REASON_CODE_REFUND = 'Refund'

IPN_PAYMENT_PERIOD_NO_PERIOD_SPECIFIED = 'NO_PERIOD_SPECIFIED'
IPN_PAYMENT_PERIOD_DAILY = 'DAILY'
IPN_PAYMENT_PERIOD_WEEKLY = 'WEEKLY'
IPN_PAYMENT_PERIOD_BIWEEKLY = 'BIWEEKLY'
IPN_PAYMENT_PERIOD_SEMIMONTHLY = 'SEMIMONTHLY'
IPN_PAYMENT_PERIOD_MONTHLY = 'MONTHLY'
IPN_PAYMENT_PERIOD_ANNUALLY = 'ANNUALLY'

IPN_PIN_TYPE_NOT_REQUIRED = 'NOT_REQUIRED'
IPN_PIN_TYPE_REQUIRED = 'REQUIRED'

IPN_TIMEZONES = {'PDT': timezone('US/Pacific'),
                 'PST': timezone('US/Pacific')}


from celery.task import task
from celery.utils.log import get_task_logger

from .models import Preapproval, Payment

logger = get_task_logger(__name__)


@task
def update_preapproval(preapproval_id):
    preapproval = Preapproval.objects.get(pk=preapproval_id)
    if preapproval.status != 'used':
        logger.info('Updating Preapproval %i' % preapproval.id)
        preapproval.update()


@task
def update_payment(payment_id):
    payment = Payment.objects.get(pk=payment_id)
    if payment.status != 'completed':
        logger.info('Updating Payment %i' % payment.id)
        payment.update()



from django.test import TestCase
from django.test import Client

from paypaladaptive.models import Preapproval

from factories import PreapprovalFactory


class TestPreapprovalReturnURL(TestCase):
    def setUp(self):
        self.preapproval = PreapprovalFactory.create(status='created')

    def get_preapproval(self):
        return Preapproval.objects.get(pk=self.preapproval.pk)

    def hitReturnUrl(self):
        data = {'next': '/'}
        client = Client()
        return client.get(self.preapproval.return_url, data=data, follow=False)

    def testPassing(self):
        """Test a normal call does what it should"""

        self.preapproval.status = 'created'
        self.preapproval.save()

        response = self.hitReturnUrl()
        # accept normal response or redirect
        self.assertIn(response.status_code, [200, 302])

        preapproval = self.get_preapproval()

        self.assertEqual(preapproval.status_detail, '')
        self.assertEqual(preapproval.status, 'returned')

    def testStatusAlreadyApproved(self):
        """
        Test for the case the IPN is received before the return url is accessed
        by the user. That is, the status is approved.

        """

        self.preapproval.status = 'approved'
        self.preapproval.save()

        response = self.hitReturnUrl()
        # accept normal response or redirect
        self.assertIn(response.status_code, [200, 302])

        preapproval = self.get_preapproval()
        self.assertEqual(preapproval.status_detail, '')
        self.assertEqual(preapproval.status, 'approved')

    def testBadStatus(self):
        """Test with bad saved Preapproval status"""

        self.preapproval.status = 'canceled'
        self.preapproval.save()

        response = self.hitReturnUrl()

        self.assertEqual(response.status_code, 500)

        p = self.get_preapproval()
        self.assertEqual(p.status, 'error')
        self.assertEqual(p.status_detail,
                        "Expected status to be created or approved not %s - "
                        "duplicate transaction?" % 'canceled')



from datetime import timedelta
from django.conf import settings
from money import set_default_currency


DEBUG = getattr(settings, "DEBUG", False)

if DEBUG:
    # use sandboxes while in debug mode
    PAYPAL_ENDPOINT = 'https://svcs.sandbox.paypal.com/AdaptivePayments/'
    PAYPAL_PAYMENT_HOST = 'https://www.sandbox.paypal.com/au/cgi-bin/webscr'
    EMBEDDED_ENDPOINT = 'https://www.sandbox.paypal.com/webapps/adaptivepayment/flow/pay'

    PAYPAL_APPLICATION_ID = 'APP-80W284485P519543T' # sandbox only
else:
    PAYPAL_ENDPOINT = 'https://svcs.paypal.com/AdaptivePayments/' # production
    PAYPAL_PAYMENT_HOST = 'https://www.paypal.com/webscr' # production
    EMBEDDED_ENDPOINT = 'https://paypal.com/webapps/adaptivepayment/flow/pay'

    PAYPAL_APPLICATION_ID = settings.PAYPAL_APPLICATION_ID

# These settings are required
PAYPAL_USERID = settings.PAYPAL_USERID
PAYPAL_PASSWORD = settings.PAYPAL_PASSWORD
PAYPAL_SIGNATURE = settings.PAYPAL_SIGNATURE
PAYPAL_EMAIL = settings.PAYPAL_EMAIL

USE_IPN = getattr(settings, 'PAYPAL_USE_IPN', True)
USE_DELAYED_UPDATES = getattr(settings, 'PAYPAL_USE_DELAYED_UPDATES', False)
DELAYED_UPDATE_COUNTDOWN = getattr(
    settings, 'PAYPAL_DELAYED_UPDATE_COUNTDOWN', timedelta(minutes=60))
USE_CHAIN = getattr(settings, 'PAYPAL_USE_CHAIN', True)
USE_EMBEDDED = getattr(settings, 'PAYPAL_USE_EMBEDDED', True)
SHIPPING = getattr(settings, 'PAYPAL_USE_SHIPPING', False)

DEFAULT_CURRENCY = getattr(settings, 'DEFAULT_CURRENCY', 'USD')
set_default_currency(code=DEFAULT_CURRENCY)

DECIMAL_PLACES = getattr(settings, 'PAYPAL_DECIMAL_PLACES', 2)
MAX_DIGITS = getattr(settings, 'PAYPAL_MAX_DIGITS', 10)

# Should tests hit Paypaladaptive or not? Defaults to using mock responses
TEST_WITH_MOCK = getattr(settings, 'PAYPAL_TEST_WITH_MOCK', True)



from tests import AdaptiveTests
from ipn import TestPaymentIPN, TestPreapprovalIPN, TestIPNVerification
from preapproval_return_url import TestPreapprovalReturnURL
from preapproval_cancel import TestPreapprovalCancel
from preapproval_update import TestPreapprovalUpdate
from payment import PaymentDBTest
from payment_return_url import TestPaymentReturnURL
from payment_response import TestPaymentResponses
from payment_update import TestPaymentUpdate



import json

from django.test import TestCase

from mock import patch

from .factories import PaymentFactory


class MockUpdateRequest(object):
    _response = None
    _code = None
    _base_response = {
        'responseEnvelope': {
            'ack': 'Success'
        }
    }

    @property
    def response(self):
        return self._response

    @property
    def code(self):
        return self._code

    def call(self, *args, **kwargs):
        return self

    @classmethod
    def set_response(cls, response):
        response.update(cls._base_response)
        cls._response = json.dumps(response)


class TestPaymentUpdate(TestCase):
    def setUp(self):
        self.payment = PaymentFactory.create(
            status='new', pay_key='AP-9HW83863H61516232')

    def test_parse_update_status(self):
        response = {'status': 'COMPLETED'}
        self.assertEqual(
            'completed', self.payment._parse_update_status(response))

        response = {'status': 'CREATED'}
        self.assertEqual(
            'created', self.payment._parse_update_status(response))

        response = {'status': 'ERROR'}
        self.assertEqual(
            'error', self.payment._parse_update_status(response))

        self.assertEqual('new', self.payment._parse_update_status({}))

    def test_get_update_kwargs(self):
        self.assertEqual({'payKey': self.payment.pay_key},
                         self.payment.get_update_kwargs())

        with self.assertRaises(ValueError) as context_manager:
            self.payment.pay_key = None
            self.payment.get_update_kwargs()

        self.assertEqual(context_manager.exception.message,
                         "Can't update unprocessed payments")

    @patch("paypaladaptive.api.endpoints.UrlRequest", MockUpdateRequest)
    def test_update(self):
        MockUpdateRequest.set_response({'status': 'COMPLETED'})

        self.payment.status = 'created'
        self.payment.update()
        self.assertEqual(self.payment.status, 'completed')

        MockUpdateRequest.set_response({'status': 'CREATED'})

        self.payment.update()
        self.assertEqual(self.payment.status, 'created')

        MockUpdateRequest.set_response({'status': 'ERROR'})

        self.payment.update()
        self.assertEqual(self.payment.status, 'error')

        self.payment.status = 'new'

        MockUpdateRequest.set_response({
            'status': 'an unrecognized weird value'})

        self.payment.update()
        self.assertEqual(self.payment.status, 'new')




import logging

from django.http import HttpResponseBadRequest

from api.ipn import IPN
from api import IpnError

logger = logging.getLogger(__name__)

def takes_ipn(function):
    def _view(request, *args, **kwargs):
        try:
            kwargs['ipn'] = IPN(request)
        except IpnError, e:
            logger.warning("PayPal IPN verify failed: %s" % e)
            logger.debug("Request was: %s" % request)
            return HttpResponseBadRequest('verify failed')

        logger.debug("Incoming IPN call: " + str(request))

        return function(request, *args, **kwargs)

    return _view


import json

from django.test import TestCase

from mock import patch

from .factories import PreapprovalFactory


class MockUpdateRequest(object):
    _response = None
    _code = None
    _base_response = {
        'responseEnvelope': {
            'ack': 'Success'
        }
    }

    @property
    def response(self):
        return self._response

    @property
    def code(self):
        return self._code

    def call(self, *args, **kwargs):
        return self

    @classmethod
    def set_response(cls, response):
        response.update(cls._base_response)
        cls._response = json.dumps(response)


class TestPreapprovalUpdate(TestCase):
    def setUp(self):
        self.preapproval = PreapprovalFactory.create(
            status='new', preapproval_key='PA-9HW83863H61516232')

    def test_parse_status(self):
        response = {'curPayments': 1,
                    'maxNumberOfPayments': 1}
        self.assertEqual(
            'used', self.preapproval._parse_update_status(response))
        
        response = {'status': 'ACTIVE', 'approved': 'true'}
        self.assertEqual(
            'approved', self.preapproval._parse_update_status(response))

        response = {'status': 'ACTIVE', 'approved': 'false'}
        self.assertEqual(
            'created', self.preapproval._parse_update_status(response))

        response = {'status': 'CANCELED'}
        self.assertEqual(
            'canceled', self.preapproval._parse_update_status(response))

        self.assertEqual('new', self.preapproval._parse_update_status({}))

    def test_get_update_kwargs(self):
        self.assertEqual({'preapprovalKey': self.preapproval.preapproval_key},
                         self.preapproval.get_update_kwargs())

        with self.assertRaises(ValueError) as context_manager:
            self.preapproval.preapproval_key = None
            self.preapproval.get_update_kwargs()

        self.assertEqual(context_manager.exception.message,
                         "Can't update unprocessed preapprovals")

    @patch("paypaladaptive.api.endpoints.UrlRequest", MockUpdateRequest)
    def test_update(self):
        MockUpdateRequest.set_response({
            'curPayments': 1,
            'maxNumberOfPayments': 1,
            'status': 'ACTIVE',
            'approved': 'true'
        })

        self.preapproval.status = 'created'
        self.preapproval.update()
        self.assertEqual(self.preapproval.status, 'used')

        MockUpdateRequest.set_response({
            'curPayments': 0,
            'maxNumberOfPayments': 1,
            'status': 'ACTIVE',
            'approved': 'true'
        })

        self.preapproval.update()
        self.assertEqual(self.preapproval.status, 'approved')

        MockUpdateRequest.set_response({
            'curPayments': 0,
            'maxNumberOfPayments': 1,
            'status': 'ACTIVE',
            'approved': 'false'
        })

        self.preapproval.update()
        self.assertEqual(self.preapproval.status, 'created')

        MockUpdateRequest.set_response({
            'curPayments': 0,
            'maxNumberOfPayments': 1,
            'status': 'CANCELED',
            'approved': 'true'
        })

        self.preapproval.update()
        self.assertEqual(self.preapproval.status, 'canceled')




from django.contrib import admin


def update_adaptive_instance(modeladmin, request, queryset):
    for instance in queryset:
        instance.update(save=True)

update_adaptive_instance.short_description = u"Update"


class PreapprovalAdmin(admin.ModelAdmin):
    list_display = ('preapproval_key', 'valid_until_date', 'status')
    actions = [update_adaptive_instance]


class PaymentAdmin(admin.ModelAdmin):
    actions = [update_adaptive_instance]


class RefundAdmin(admin.ModelAdmin):
    pass



"""
Paypal Adaptive Payments supporting views

"""

import logging

from django.contrib.auth.decorators import login_required
from django.db import transaction
from django.http import (HttpResponseServerError, HttpResponseRedirect,
                         HttpResponseBadRequest, HttpResponse, Http404)
from django.shortcuts import render_to_response
from django.template.context import RequestContext
from django.shortcuts import get_object_or_404
from django.utils.translation import ugettext_lazy as _
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST

import settings
from api.ipn import constants
from models import Payment, Preapproval
from decorators import takes_ipn


logger = logging.getLogger(__name__)


def render(request, template, template_vars=None):
    if template_vars is None:
        template_vars = {}

    if request.GET.get('next'):
        return HttpResponseRedirect(request.GET.get('next'))

    context = RequestContext(request)
    d = {"is_embedded": settings.USE_EMBEDDED}.update(template_vars)

    return render_to_response(template, d, context)


@login_required
@transaction.autocommit
def payment_cancel(request, payment_id, secret_uuid,
                   template="paypaladaptive/cancel.html"):
    """Handle incoming cancellation from paypal"""

    logger.debug("Cancellation received for Payment %s" % payment_id)

    payment = get_object_or_404(Payment, id=payment_id,
                                secret_uuid=secret_uuid)
    
    payment.status = 'canceled'
    payment.save()

    template_vars = {"is_embedded": settings.USE_EMBEDDED}
    return render(request, template, template_vars)


@transaction.autocommit
def payment_return(request, payment_id, secret_uuid,
                   template="paypaladaptive/return.html"):
    """
    Incoming return from Paypal process. Note that this is a user returning to
    the site and not a returned payment.

    """

    logger.debug("Return received for Payment %s" % payment_id)

    payment = get_object_or_404(Payment, id=payment_id,
                                secret_uuid=secret_uuid)

    if payment.status not in ['created', 'completed']:
        payment.status_detail = _(u"Expected status to be created or "
                                  u"completed, not %s - duplicate "
                                  u"transaction?") % payment.status
        payment.status = 'error'
        payment.save()
        return HttpResponseServerError('Unexpected error')

    elif secret_uuid != payment.secret_uuid:
        payment.status_detail = (_(u"BuyReturn secret \"%s\" did not match")
                                 % secret_uuid)
        payment.status = 'error'
        payment.save()
        return HttpResponseServerError('Unexpected error')

    if payment.status != 'completed':
        payment.status = 'returned'
        payment.save()

    if settings.USE_DELAYED_UPDATES:
        from .tasks import update_payment
        update_payment.delay(payment_id=payment.id)

    template_vars = {"is_embedded": settings.USE_EMBEDDED}
    return render(request, template, template_vars)


@transaction.autocommit
def preapproval_cancel(request, preapproval_id,
                       template="paypaladaptive/cancel.html"):
    """Incoming preapproval cancellation from paypal"""

    logger.debug("Cancellation received for Preapproval %s" % preapproval_id)

    get_object_or_404(Preapproval, id=preapproval_id)

    template_vars = {"is_embedded": settings.USE_EMBEDDED}
    return render(request, template, template_vars)


@transaction.autocommit
def preapproval_return(request, preapproval_id, secret_uuid,
                       template="paypaladaptive/return.html"):
    """
    Incoming return from paypal process (note this is a return to the site,
    not a returned payment)

    """

    preapproval = get_object_or_404(Preapproval, id=preapproval_id)

    logger.info("Return received for Preapproval %s" % preapproval_id)

    if preapproval.status not in ['created', 'approved']:
        preapproval.status_detail = _(
            u"Expected status to be created or approved not %s - duplicate "
            u"transaction?") % preapproval.status
        preapproval.status = 'error'
        preapproval.save()
        return HttpResponseServerError('Unexpected error')

    elif secret_uuid != preapproval.secret_uuid:
        preapproval.status_detail = _(u"BuyReturn secret \"%s\" did not"
                                      u" match") % secret_uuid
        preapproval.status = 'error'
        preapproval.save()
        return HttpResponseServerError('Unexpected error')

    if preapproval.status != 'approved':
        preapproval.status = 'returned'
        preapproval.save()

    if settings.USE_DELAYED_UPDATES:
        from .tasks import update_preapproval
        update_preapproval.delay(preapproval_id=preapproval.id)

    template_vars = {"is_embedded": settings.USE_EMBEDDED,
                     "preapproval": preapproval, }
    return render(request, template, template_vars)


@csrf_exempt
@require_POST
@transaction.autocommit
@takes_ipn
def ipn(request, object_id, object_secret_uuid, ipn):
    """
    Incoming IPN POST request from Paypal

    """

    logger.debug("Incoming IPN call: " + str(request))

    object_class = {
        constants.IPN_TYPE_PAYMENT: Payment,
        constants.IPN_TYPE_PREAPPROVAL: Preapproval,
        constants.IPN_TYPE_ADJUSTMENT: Payment
    }[ipn.type]

    try:
        obj = object_class.objects.get(pk=object_id)
    except object_class.DoesNotExist:
        logger.warning('Could not find %s ID %s, replying to IPN with '
                       '404.' % (object_class.__name__, object_id))
        raise Http404

    if obj.secret_uuid != object_secret_uuid:
        obj.status = 'error'
        obj.status_detail = ('IPN secret "%s" did not match db'
                             % object_secret_uuid)
        obj.save()
        return HttpResponseBadRequest('secret uuid mismatch')

    # IPN type-specific operations
    if ipn.type == constants.IPN_TYPE_PAYMENT:
        obj.transaction_id = ipn.transactions[0].id

        if obj.money != ipn.transactions[0].amount:
            obj.status = 'error'
            obj.status_detail = ("IPN amounts didn't match. Payment requested "
                                 "%s. Payment made %s"
                                 % (obj.money, ipn.transactions[0].amount))

        # check payment status
        elif request.POST.get('status', '') != 'COMPLETED':
            obj.status = 'error'
            obj.status_detail = ('PayPal status was "%s"'
                                 % request.POST.get('status'))
        else:
            obj.status = 'completed'

            # TODO: mark preapproval 'used'
    elif ipn.type == constants.IPN_TYPE_PREAPPROVAL:
        if obj.money != ipn.max_total_amount_of_all_payments:
            obj.status = 'error'
            obj.status_detail = (
                "IPN amounts didn't match. Preapproval requested %s. "
                "Preapproval made %s"
                % (obj.money, ipn.max_total_amount_of_all_payments))
        elif ipn.status == constants.IPN_STATUS_CANCELED:
            obj.status = 'canceled'
            obj.status_detail = 'Cancellation received via IPN'
        elif not ipn.approved:
            obj.status = 'error'
            obj.status_detail = "The preapproval is not approved"
        else:
            obj.status = 'approved'
    else:
        logger.warning(
            'No action found for IPN Type "%s" with status "%s" (id: "%s", '
            'secret_uuid: %s)'
            % (ipn.type, ipn.status, obj.id, obj.secret_uuid))

    obj.save()

    # Ok, no content
    return HttpResponse(status=204)



#!/usr/bin/env python

from setuptools import setup, find_packages
import paypaladaptive

setup(
    name='django-paypal-adaptive',
    version=".".join(map(str, paypaladaptive.__version__)),
    author='Greg McGuire',
    author_email='greg@buzzcar.com',
    maintainer='Anton Agestam',
    maintainer_email="msn@antonagestam.se",
    url='http://github.com/FundedByMe/django-paypal-adaptive',
    install_requires=[
        'Django>=1.4.3',
        'python-dateutil==2.1',
        'python-money',
    ],
    extras_require={
        'delayed-updates': ['celery>=3.0.12'],
    },
    dependency_links=[
        '-e git+http://github.com/poswald/python-money.git@29d3671e140307b958b51a32e6d1ac8553edc9e5#egg=python_money-dev'
    ],
    description='A pluggable Django application for integrating PayPal '
                'Adaptive Payments',
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        "Framework :: Django",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Operating System :: OS Independent",
        "Topic :: Software Development"
    ],
)



"""
Paypal Adaptive Payment callback and IPN URLs

"""

from django.conf.urls import patterns, url

import views
import settings


urlpatterns = patterns(
    '',

    url(r'^pay/cancel/(?P<payment_id>\d+)/(?P<secret_uuid>\w+)/$',
        views.payment_cancel, name="paypal-adaptive-payment-cancel"),

    url(r'^pay/return/(?P<payment_id>\d+)/(?P<secret_uuid>\w+)/$',
        views.payment_return, name="paypal-adaptive-payment-return"),

    url(r'^pre/cancel/(?P<preapproval_id>\d+)/$', views.preapproval_cancel,
        name="paypal-adaptive-preapproval-cancel"),

    url(r'^pre/return/(?P<preapproval_id>\d+)/(?P<secret_uuid>\w+)/$',
        views.preapproval_return, name="paypal-adaptive-preapproval-return"),
)

if settings.USE_IPN:
    urlpatterns += patterns(
        '',

        url(r'^ipn/(?P<object_id>\d+)/(?P<object_secret_uuid>\w+)/$',
            views.ipn, name="paypal-adaptive-ipn"),
    )



class PaypalAdaptiveApiError(RuntimeError):
    pass


class PayError(PaypalAdaptiveApiError):
    pass


class RefundError(PaypalAdaptiveApiError):
    pass


class CancelPreapprovalError(PaypalAdaptiveApiError):
    pass


class PreapproveError(PaypalAdaptiveApiError):
    pass


class IpnError(PaypalAdaptiveApiError):
    pass


class ReceiverError(PaypalAdaptiveApiError):
    pass


from paypaladaptive.models import Payment, Preapproval, PaypalAdaptive
import factory
import uuid
from money.Money import Money
import datetime


class PaypalAdaptiveFactory(factory.DjangoModelFactory):
    FACTORY_FOR = PaypalAdaptive

    money = Money(1400, 'SEK')
    money_currency = 'SEK'
    created_date = datetime.datetime.now()


class PaymentFactory(PaypalAdaptiveFactory):
    FACTORY_FOR = Payment
    transaction_id = str(uuid.uuid4())[0:17]


class PreapprovalFactory(PaypalAdaptiveFactory):
    FACTORY_FOR = Preapproval


from django.test import TestCase

from money.Money import Money

from ..models import Payment
from .factories import PaymentFactory


class PaymentDBTest(TestCase):

    def testSaveBigAmount(self):
        big_money = Money(2500000, "USD")
        payment = PaymentFactory.build(money=big_money,
                                       money_currency=big_money.currency)
        payment.save()

        self.assertEqual(big_money,
                         Payment.objects.get(pk=payment.pk).money)
